package org.openxma.addons.ui.table.customizer.xma.client;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.openxma.addons.ui.table.customizer.common.CommonColumnModel;
import org.openxma.addons.ui.table.customizer.common.CommonDataModel.FilterOperator;
import org.openxma.addons.ui.table.customizer.common.CommonFilterModel;

import at.spardat.xma.mdl.Atom;
import at.spardat.xma.page.IDialogPage;
import at.spardat.xma.page.PageClient;

/**
 * 
 * Helper class that automatically builds the contextmenu for the customizertable and register required listeners.
 * The contextmenu changes depending on the context - has the header been clicked or an item, was there a filter present etc.
 * May be extended to further customize the menu (ie. add custom entries). 
 * 
 * @see #buildMenu()
 * @see #addMenuItem(String, Image)
 * @see #menuItemSelectedImpl(MenuItem)
 */
public class ContextMenuHelper {
	
    private TableCustomizerComp customizer;
    private final Table table;
    private Menu menu;
    private ResourceBundle messages;
    private Locale locale;
    private int idxItemUI;
    private int idxColUI;
    private final PageClient page;
    private ContextMenuListener listener;
    private int idxColModel;
    private int idxItemModel;
    
    /**
     * 
     * 
     * @param page
     * @param table
     * @param customizer
     */
	public ContextMenuHelper(PageClient page, Table table, TableCustomizerComp customizer) {
		this.page = page;
	    this.table = table;
		this.customizer = customizer;
		
        listener = new ContextMenuListener(); //create the menu listener
		table.addListener(SWT.MenuDetect, listener);
		table.addListener(SWT.Dispose, listener);
	}

	/**
	 * Builds the context menu for the table containing default entries supported by the underlying tablecustomizer.
	 * Called every time before the menu is displayed to dynamically adjust to the currently selected item and column as well as the customizer settings.
	 * Items are appended by using the buildMenuItem method which registers the items to the listener.
	 * Extend this method to add your own custom entries before or after the default entries are added and implement {@link #menuItemSelectedImpl(MenuItem)} to react to selection.
	 * 
	 * @see #addMenuItem(String, Image)
	 * @see #addMenuSeparator()
	 * @see #menuItemSelectedImpl(MenuItem)
	 */
    protected void buildMenu() {
        if (idxColUI != -1) {
            // Filter & Copy Area
            CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(idxColModel);
            if (idxItemUI != -1) {
                String text = getExternalValue();
                if (text != null && text.length() > 0) {
                    addMenuItem(getMessage("CopyToClipboard"), null);
                }
                if (columnModel != null) { //cannot filter without columnmodel
                    if (text != null && text.length() > 40) { //trim displayed value if too long.
                        text = text.substring(0, 37).concat("...");
                    }
                    addMenuItem(getMessage("FilterBy")+" \""+text+"\"", customizer.getImage(TableCustomizerComp.IMG_FILTER)); // could be usefull to search for "" or null?
                }
            } else if (columnModel != null){
                addMenuItem(getMessage("FilterDots"), customizer.getImage(TableCustomizerComp.IMG_FILTER));
            }
            if (columnModel != null && !columnModel.getFilterModel().isEmpty()) {
                addMenuItem(getMessage("DeleteFilter"), customizer.getImage(TableCustomizerComp.IMG_FILTER_DELETE));
            }
            
            // Sort Area
            addMenuSeparator();
            if (columnModel != null) {
                if (columnModel.getSortOrder() != 0 || !columnModel.isAscending()) {
                    addMenuItem(getMessage("SortAsc"), customizer.getImage(TableCustomizerComp.IMG_SORT_ASC));
                }
                if (columnModel.getSortOrder() != 0 || !columnModel.isDescending()) {
                    addMenuItem(getMessage("SortDesc"), customizer.getImage(TableCustomizerComp.IMG_SORT_DESC));
                }
                if (columnModel.getSortOrder() != Integer.MAX_VALUE) {
                    addMenuItem(getMessage("SortRemove"), null);
                }
            } else { //Sortmenu without customizer support
                if (customizer.getPagingWMClient().getSortingColumn() != idxColModel || !customizer.getPagingWMClient().getAscending()) {
                    addMenuItem(getMessage("SortAsc"), customizer.getImage(TableCustomizerComp.IMG_SORT_ASC));
                }
                if (customizer.getPagingWMClient().getSortingColumn() != idxColModel || customizer.getPagingWMClient().getAscending()) {
                    addMenuItem(getMessage("SortDesc"), customizer.getImage(TableCustomizerComp.IMG_SORT_DESC));
                }
                if (customizer.getPagingWMClient().getSortingColumn() == idxColModel) {
                    addMenuItem(getMessage("SortRemove"), null);
                }
            }
        }
    }

    private void setMenuData(TableColumn column, TableItem item) {
        menu.setData("COLUMN",column);
        menu.setData("ITEM",item);
        if (item != null) {
            idxItemUI = table.indexOf(item);
            idxItemModel = customizer.getTableWMClient().uiIndex2ModelIndex(idxItemUI);
        } else {
            idxItemUI = -1;
            idxItemModel = -1;
        }
        
        if (column != null) {
            idxColUI = table.indexOf(column);
            idxColModel = customizer.getTableWMClient().uiColumnIndexToModelColumnIndex(idxColUI);
        } else {
            idxColUI = -1;
            idxColModel = -1;
        }
    }

    private String getExternalValue() {
        Atom cell = extractCell();
        return cell != null ? customizer.getTableWMClient().getFormatter(idxColModel).format(cell.getEncodedValue()) : null;
    }

    private Atom extractCell() {
        return customizer.getTableWMClient().getRow(idxItemModel).getCell(idxColModel);
    }

    /**
     * Adds a separator to the current menu layout.
     * 
     * @see #buildMenu(TableColumn column, TableItem item)
     */
    protected final void addMenuSeparator() {
        if (menu.getItemCount() > 0) {
            new MenuItem(menu, SWT.SEPARATOR);
        }
    }

    /**
     * Adds a menu item to the current menu layout, registers it to the listener and returns it. 
     * 
     * @param text The text for the MenuItem entry.
     * @param image The icon for the MenuItem entry.
     * @return MenuItem The newly added menuItem for further reference.
     * 
     * @see #buildMenu(TableColumn column, TableItem item)
     * @see #widgetSelectedImpl(SelectionEvent e)
     */
    protected final MenuItem addMenuItem(String text, Image image) {
        MenuItem item = new MenuItem(menu, menu.getStyle());
        item.setText(text);
        item.setImage(image);
        item.addListener(SWT.Selection, listener);
        return item;
    }
    
    /**
     * Extend this method to implement selection behaviour for menuitems added via the {@link #addMenuItem(String, Image)} method.
     * Called everytime a menuitem in the contextmenu was selected.
     * 
     * @see #addMenuItem(String, Image)
     */
    protected void menuItemSelected(MenuItem menuItem) {
        TableColumn column = getSelectedColumn();
        String selectedText = menuItem.getText();
        if (getMessage("FilterDots").equals(selectedText)) {
            customizer.callFilterDialog(column);
        } else if (selectedText.length() > 12 && getMessage("FilterBy").equals(selectedText.substring(0, 12))) {
            CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(idxColModel);
            String externalValue = getExternalValue();
            if (externalValue != null) {
                columnModel.setFilterModel(Arrays.asList(new CommonFilterModel(FilterOperator.eq, externalValue, customizer.getTableWMClient().getColumn(idxColUI).getFormatter())));
            } else {
                columnModel.setFilterModel(Arrays.asList(new CommonFilterModel(FilterOperator.isnull, null)));
            }
            customizer.applySettings(true);
        }else if (getMessage("DeleteFilter").equals(selectedText)) {
            CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(idxColModel);
            columnModel.removeFilterModel();
            customizer.applySettings(true);
        } else if (getMessage("SortAsc").equals(selectedText)) {
            customizer.sort(idxColModel, true);
        } else if (getMessage("SortDesc").equals(selectedText)) {
            customizer.sort(idxColModel, false);
        }  else if (getMessage("SortRemove").equals(selectedText)) {
            customizer.sort(idxColModel, null);
        } else if (getMessage("CopyToClipboard").equals(selectedText)) {
            handleCopy(getSelectedItem().getText(idxColUI));
        }
    }

    /**
     * @return the TableItem which the contextmenu was built for.
     */
    protected final TableItem getSelectedItem() {
        return (TableItem) menu.getData("ITEM");
    }

    /**
     * @return the TableColumn which the contextmenu was built for.
     */
    protected final TableColumn getSelectedColumn() {
        return (TableColumn) menu.getData("COLUMN");
    }

    /**
     * @return the key of the item which the contextmenu was built for.
     */
    protected final String getMenuItemKey() {
        return customizer.getTableWMClient().getRow(idxItemUI).getKey();
    }

    private void handleCopy(String copyMenuClipboardText_) {
        if (copyMenuClipboardText_ != null && copyMenuClipboardText_.length()>0) {
            Display display = table.getShell().getDisplay();
            Clipboard clipboard = new Clipboard(display);
            clipboard.setContents(new String[] {copyMenuClipboardText_}, new Transfer[] {TextTransfer.getInstance()});
            clipboard.dispose();
        }
    }

    private String getMessage(String key) {
        if (messages == null) {
            messages = ResourceBundle.getBundle("org.openxma.addons.ui.table.customizer.xma.client.ContextMenuHelper",getLocale());
        }
        return messages.getString(key);
    }
    
    private Locale getLocale() {
        if (locale == null) {
            locale = customizer.getContext().getLocale();
        }
        return locale;
    }
    
    private class ContextMenuListener implements Listener {
        
        public ContextMenuListener() {}

        public void handleEvent(Event event) {
            if (!page.isEventsEnabled()) return;
            IDialogPage dialog = page.getDialogPage();
            try {
                dialog.setEventsEnabled(false);
                if (event.type == SWT.MenuDetect) {
                    handleMenuDetect(event);
                } else if (event.type == SWT.Selection){
                    handleMenuItemSelect(event);
                } else if (event.type == SWT.Dispose){
                    handleMenuDispose();
                }
            } catch (Exception exc) {
                page.showException(exc);
            } finally {
                dialog.setEventsEnabled(true);
            }
        }

        private void handleMenuDetect(Event event) {
            Point pt = table.getDisplay().map(null, table, new Point(event.x, event.y));
            TableColumn column = getColumnByPoint(pt);
            TableItem item = table.getItem(pt);
            menu = new Menu(table);
            setMenuData(column, item);
            buildMenu();
            table.setMenu(menu);
        }

        private void handleMenuItemSelect(Event event) {
            if (event.widget instanceof MenuItem) {
                menuItemSelected((MenuItem) event.widget);
            }
        }

        private void handleMenuDispose() {
            if (menu != null && !menu.isDisposed()) {
                menu.dispose();
            }
        }
        
        private TableColumn getColumnByPoint(Point pt) {
            int x = 0;
            if (table.getHorizontalBar() != null) { // factor in possible horizontal bar offset
                x -= table.getHorizontalBar().getSelection();
            }
            
            List<Integer> fullColumnOrder = getFullColumnOrder();
            for (int order : fullColumnOrder) {
                TableColumn column = table.getColumn(customizer.getTableWMClient().modelColumnIndexToUIColumnIndex(order));
                x += column.getWidth();
                if (pt.x <= x) {
                    return column;
                }
            }
            return null;
        }
        
        //Fill with columns that are not in the tablemodel
        private List<Integer> getFullColumnOrder() {
            List<Integer> fullColumnOrder = new ArrayList<Integer>(customizer.getTableModel().getColumnOrder().length);
            for (int i : customizer.getTableModel().getColumnOrder()) {
                fullColumnOrder.add(i);
            }
            for (int i = 0; i < customizer.getTableWMClient().getColumnCount();i++) {
                if (!fullColumnOrder.contains(i)) {
                    fullColumnOrder.add(i, i);
                }
            }
            return fullColumnOrder;
        }
    }
}